@@ -31,7 +31,7 @@ import apiritif
 from apiritif.ssl_adapter import SSLAdapter
 from apiritif.thread import get_from_thread_store, put_into_thread_store
 from apiritif.utilities import *
-from apiritif.utils import headers_as_text, assert_regexp, assert_not_regexp, log, get_trace
+from apiritif.utils import headers_as_text, assert_regexp, assert_not_regexp, log, get_trace, NormalShutdown, graceful
 
 BODY_LIMIT = int(os.environ.get("APIRITIF_TRACE_BODY_EXCLIMIT", "1024"))
 
@@ -221,14 +221,14 @@ class smart_transaction(transaction_logged):
         self.driver, self.func_mode, self.controller = get_from_thread_store(("driver", "func_mode", "controller"))
 
         if self.controller.tran_mode:
-            self.controller.beforeTest()
+            self.controller.startTest()
         else:
             self.controller.tran_mode = True
 
         self.test_suite = self.controller.current_sample.test_suite
 
     def __enter__(self):
-        self.controller.startTest()
+        self.controller.set_start_time()
 
         super(smart_transaction, self).__enter__()
         put_into_thread_store(test_case=self.name, test_suite=self.test_suite)
@@ -237,7 +237,6 @@ class smart_transaction(transaction_logged):
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         super(smart_transaction, self).__exit__(exc_type, exc_val, exc_tb)
-        self.controller.stopTest(is_transaction=True)
 
         message = ''
 
@@ -260,15 +259,13 @@ class smart_transaction(transaction_logged):
         for func in apiritif.get_transaction_handlers()["exit"]:
             func()
 
-        self.controller.afterTest(is_transaction=True)
+        self.controller.stopTest(is_transaction=True)
 
-        graceful_file_name = os.environ.get('GRACEFUL')
-        graceful_flag = graceful_file_name and os.path.exists(graceful_file_name)
         stage = apiritif.get_stage()
         if stage == "teardown":
             self.func_mode = False
-        elif graceful_flag:     # and stage in ("setup", "main")
-            raise StopIteration("graceful!")
+        elif graceful():     # and stage in ("setup", "main")
+            raise NormalShutdown("graceful!")
 
         return not self.func_mode  # don't reraise in load mode
 
